const extend = require('../../../core/utils/extend').extend;
const _extend = extend;

const symbolPoint = require('./symbol_point');
const barPoint = require('./bar_point');
const piePoint = require('./pie_point');
const isDefined = require('../../../core/utils/type').isDefined;
const vizUtils = require('../../core/utils');
const normalizeAngle = vizUtils.normalizeAngle;

const _math = Math;
const _max = _math.max;

const RADIAL_LABEL_INDENT = require('../../components/consts').radialLabelIndent;

const ERROR_BARS_ANGLE_OFFSET = 90;
const CANVAS_POSITION_END = 'canvas_position_end';
const CANVAS_POSITION_DEFAULT = 'canvas_position_default';

exports.polarSymbolPoint = _extend({}, symbolPoint, {

    _getLabelCoords: piePoint._getLabelCoords,

    _moveLabelOnCanvas: barPoint._moveLabelOnCanvas,

    _getLabelPosition: function() {
        return 'outside';
    },

    _getCoords: function(argument, value) {
        const axis = this.series.getValueAxis();
        const startAngle = axis.getAngles()[0];
        const angle = this._getArgTranslator().translate(argument);
        const radius = this._getValTranslator().translate(value);
        const coords = vizUtils.convertPolarToXY(axis.getCenter(), axis.getAngles()[0], angle, radius);

        coords.angle = angle + startAngle - 90,
        coords.radius = radius;

        return coords;
    },

    _translate: function() {
        const that = this;
        const center = that.series.getValueAxis().getCenter();
        const coord = that._getCoords(that.argument, that.value);

        that.vx = normalizeAngle(coord.angle);
        that.vy = that.radiusOuter = that.radiusLabels = coord.radius;
        that.radiusLabels += RADIAL_LABEL_INDENT;

        that.radius = coord.radius;
        that.middleAngle = -coord.angle;
        that.angle = -coord.angle;

        that.x = coord.x;
        that.y = coord.y;
        that.defaultX = that.centerX = center.x;
        that.defaultY = that.centerY = center.y;

        that._translateErrorBars();

        that.inVisibleArea = true;
    },

    _translateErrorBars: function() {
        const that = this;
        const errorBars = that._options.errorBars;
        const translator = that._getValTranslator();

        if(!errorBars) {
            return;
        }

        isDefined(that.lowError) && (that._lowErrorCoord = that.centerY - translator.translate(that.lowError));
        isDefined(that.highError) && (that._highErrorCoord = that.centerY - translator.translate(that.highError));
        that._errorBarPos = that.centerX;

        that._baseErrorBarPos = errorBars.type === 'stdDeviation' ? that._lowErrorCoord + (that._highErrorCoord - that._lowErrorCoord) / 2 : that.centerY - that.radius;
    },

    _getTranslates: function(animationEnabled) {
        return animationEnabled ? this.getDefaultCoords() : { x: this.x, y: this.y };
    },

    getDefaultCoords: function() {
        const cosSin = vizUtils.getCosAndSin(-this.angle);
        const radius = this._getValTranslator().translate(CANVAS_POSITION_DEFAULT);
        const x = this.defaultX + radius * cosSin.cos;
        const y = this.defaultY + radius * cosSin.sin;

        return { x: x, y: y };
    },

    _addLabelAlignmentAndOffset: function(label, coord) {
        return coord;
    },

    _checkLabelPosition: function(label, coord) {
        const that = this;
        const visibleArea = that._getVisibleArea();
        const graphicBBox = that._getGraphicBBox();

        if(that._isPointInVisibleArea(visibleArea, graphicBBox)) {
            coord = that._moveLabelOnCanvas(coord, visibleArea, label.getBoundingRect());
        }

        return coord;
    },

    _getErrorBarSettings: function(errorBarOptions, animationEnabled) {
        const settings = symbolPoint._getErrorBarSettings.call(this, errorBarOptions, animationEnabled);

        settings.rotate = ERROR_BARS_ANGLE_OFFSET - this.angle;
        settings.rotateX = this.centerX;
        settings.rotateY = this.centerY;

        return settings;
    },

    getCoords: function(min) {
        return min ? this.getDefaultCoords() : { x: this.x, y: this.y };
    }
});

exports.polarBarPoint = _extend({}, barPoint, {

    _translateErrorBars: exports.polarSymbolPoint._translateErrorBars,

    _getErrorBarSettings: exports.polarSymbolPoint._getErrorBarSettings,

    _moveLabelOnCanvas: barPoint._moveLabelOnCanvas,

    _getLabelCoords: piePoint._getLabelCoords,

    _getLabelConnector: piePoint._getLabelConnector,

    getTooltipParams: piePoint.getTooltipParams,

    _getLabelPosition: piePoint._getLabelPosition,

    _getCoords: exports.polarSymbolPoint._getCoords,

    _translate: function() {
        const that = this;
        const translator = that._getValTranslator();
        const maxRadius = translator.translate(CANVAS_POSITION_END);

        that.radiusInner = translator.translate(that.minValue);

        exports.polarSymbolPoint._translate.call(that);

        if(that.radiusInner === null) {
            that.radiusInner = that.radius = maxRadius;
        } else if(that.radius === null) {
            this.radius = this.value >= 0 ? maxRadius : 0;
        }

        that.radiusOuter = that.radiusLabels = _max(that.radiusInner, that.radius);
        that.radiusLabels += RADIAL_LABEL_INDENT;
        that.radiusInner = that.defaultRadius = _math.min(that.radiusInner, that.radius);

        that.middleAngle = that.angle = -normalizeAngle(that.middleAngleCorrection - that.angle);
    },

    _getErrorBarBaseEdgeLength() {
        const coord = this.getMarkerCoords();
        return _math.PI * coord.outerRadius * _math.abs(coord.startAngle - coord.endAngle) / 180;
    },

    getMarkerCoords: function() {
        return {
            x: this.centerX,
            y: this.centerY,
            outerRadius: this.radiusOuter,
            innerRadius: this.defaultRadius,
            startAngle: this.middleAngle - this.interval / 2,
            endAngle: this.middleAngle + this.interval / 2
        };
    },

    _drawMarker: function(renderer, group, animationEnabled) {
        const that = this;
        const styles = that._getStyle();
        const coords = that.getMarkerCoords();
        let innerRadius = coords.innerRadius;
        let outerRadius = coords.outerRadius;

        const start = that._getCoords(that.argument, CANVAS_POSITION_DEFAULT);
        let x = coords.x;
        let y = coords.y;

        if(animationEnabled) {
            innerRadius = 0;
            outerRadius = 0;
            x = start.x;
            y = start.y;
        }

        that.graphic = renderer.arc(x, y, innerRadius, outerRadius, coords.startAngle, coords.endAngle).attr(styles).data({ 'chart-data-point': that }).append(group);
    },

    _checkLabelPosition: function(label, coord) {
        const that = this;
        const visibleArea = that._getVisibleArea();
        const angleFunctions = vizUtils.getCosAndSin(that.middleAngle);
        const x = that.centerX + that.defaultRadius * angleFunctions.cos;
        const y = that.centerY - that.defaultRadius * angleFunctions.sin;

        if(x > visibleArea.minX && x < visibleArea.maxX && y > visibleArea.minY && y < visibleArea.maxY) {
            coord = that._moveLabelOnCanvas(coord, visibleArea, label.getBoundingRect());
        }
        return coord;
    },

    _addLabelAlignmentAndOffset: function(label, coord) {
        return coord;
    },

    correctCoordinates: function(correctOptions) {
        this.middleAngleCorrection = correctOptions.offset;
        this.interval = correctOptions.width;
    },

    coordsIn: function(x, y) {
        const val = vizUtils.convertXYToPolar(this.series.getValueAxis().getCenter(), x, y);
        const coords = this.getMarkerCoords();
        const isBetweenAngles = coords.startAngle < coords.endAngle ?
            -val.phi >= coords.startAngle && -val.phi <= coords.endAngle :
            -val.phi <= coords.startAngle && -val.phi >= coords.endAngle;

        return (val.r >= coords.innerRadius && val.r <= coords.outerRadius && isBetweenAngles);
    }
});
